@@ -1,19 +1,18 @@
-local helper = wesnoth.require "helper"
 local T = wml.tag
 local vars = wml.variables
 
 function wesnoth.wml_actions.find_respawn_point(cfg)
-	local respawn_near = cfg.respawn_near or helper.wml_error "[find_respawn_point] missing required respawn_near= key"
+	local respawn_near = cfg.respawn_near or wml.error "[find_respawn_point] missing required respawn_near= key"
 	local variable = cfg.variable or "nearest_hex"
 	local respawn_point
 	local radius = 1
 
-	if not wesnoth.eval_conditional { T.have_unit { id = respawn_near} } then
+	if not wml.eval_conditional { T.have_unit { id = respawn_near} } then
 		respawn_near = "Tallin"
 	end
 
 	repeat
-		respawn_point = wesnoth.get_locations({
+		respawn_point = wesnoth.map.find({
 			include_borders = false,
 			T["and"] {
 				T.filter {
